@@ -1,27 +1,37 @@
 # -*- coding: utf-8 -*-
 """Test bregex lib."""
-from __future__ import unicode_literals
 import unittest
 from backrefs import bregex
 from backrefs import _bregex_parse
 import regex
-import sys
 import pytest
-import _regex_core
 import random
 import copy
+import time
+import sys
+try:
+    import _regex_core
+except ImportError:
+    from regex import _regex_core
 
-PY3 = (3, 0) <= sys.version_info < (4, 0)
-
-if PY3:
-    binary_type = bytes  # noqa
-else:
-    binary_type = str  # noqa
+PY36_PLUS = (3, 6) <= sys.version_info
+PY39_PLUS = (3, 9) <= sys.version_info
 
 
 class TestSearchTemplate(unittest.TestCase):
     """Search template tests."""
 
+    def test_named_lists(self):
+        """Test named lists."""
+
+        kwargs = {'name': ['blue', 'green']}
+        self.assertTrue(bregex.compile(r'\L<name>', name=['blue', 'green']).match('green') is not None)
+        self.assertTrue(bregex.compile(r'\L<name>', **kwargs).match('green') is not None)
+        self.assertTrue(bregex.compile_search(r'\L<name>', name=['blue', 'green']).match('green') is not None)
+        self.assertTrue(bregex.compile_search(r'\L<name>', **kwargs).match('green') is not None)
+        self.assertTrue(bregex.match(r'\L<name>', 'green', name=['blue', 'green']) is not None)
+        self.assertTrue(bregex.match(r'\L<name>', 'green', **kwargs) is not None)
+
     def test_non_raw_string_unicode(self):
         """Test non raw string Unicode notation."""
 
@@ -57,13 +67,15 @@ class TestSearchTemplate(unittest.TestCase):
         p1 = bregex.compile('test')
         p2 = bregex.compile('test')
         p3 = bregex.compile('test', bregex.X)
+        p4 = bregex.compile(b'test')
 
         self.assertTrue(p1 == p2)
         self.assertTrue(p1 != p3)
+        self.assertTrue(p1 != p4)
 
-        p4 = copy.copy(p1)
-        self.assertTrue(p1 == p4)
-        self.assertTrue(p4 in {p1})
+        p5 = copy.copy(p1)
+        self.assertTrue(p1 == p5)
+        self.assertTrue(p5 in {p1})
 
     def test_not_flags(self):
         """Test invalid flags."""
@@ -96,6 +108,11 @@ class TestSearchTemplate(unittest.TestCase):
     def test_cache(self):
         """Test cache."""
 
+        try:
+            from regex import _cache
+        except ImportError:
+            from regex.regex import _cache
+
         bregex.purge()
         self.assertEqual(bregex._get_cache_size(), 0)
         self.assertEqual(bregex._get_cache_size(True), 0)
@@ -110,7 +127,7 @@ class TestSearchTemplate(unittest.TestCase):
         bregex.purge()
         self.assertEqual(bregex._get_cache_size(), 0)
         self.assertEqual(bregex._get_cache_size(True), 0)
-        self.assertEqual(len(regex._cache), 0)
+        self.assertEqual(len(_cache), 0)
 
     def test_infinite_loop_catch(self):
         """Test infinite loop catch."""
@@ -148,11 +165,11 @@ class TestSearchTemplate(unittest.TestCase):
 
         self.assertEqual(
             pattern.pattern,
-            r'''(?uV0)Test # (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)(?#\R\))(?x:
+            r'''(?uV0)Test # (?>\r\n|[\n\v\f\r\x85\u2028\u2029])(?#\R\))(?x:
             Test #\\R(?#\\R\)
             (Test # \\R
             )Test #\\R
-            )Test # (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)'''
+            )Test # (?>\r\n|[\n\v\f\r\x85\u2028\u2029])'''
         )
 
         self.assertTrue(pattern.match('Test # \nTestTestTestTest # \n') is not None)
@@ -163,7 +180,7 @@ class TestSearchTemplate(unittest.TestCase):
         pattern = bregex.compile_search(r'(?V1)# \R (?x) # \R')
         self.assertEqual(
             pattern.pattern,
-            r'(?V1)# (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029) (?x) # \\R'
+            r'(?V1)# (?>\r\n|[\n\v\f\r\x85\u2028\u2029]) (?x) # \\R'
         )
 
         self.assertTrue(pattern.match('# \n '))
@@ -185,9 +202,9 @@ class TestSearchTemplate(unittest.TestCase):
             pattern.pattern,
             r'''(?xuV1)
             Test # \\R
-            (?-x:Test #(?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)(?#\R\))((?x)
+            (?-x:Test #(?>\r\n|[\n\v\f\r\x85\u2028\u2029])(?#\R\))((?x)
             Test # \\R
-            )Test #(?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029))
+            )Test #(?>\r\n|[\n\v\f\r\x85\u2028\u2029]))
             Test # \\R
             '''
         )
@@ -219,16 +236,16 @@ class TestSearchTemplate(unittest.TestCase):
         """Test char group with nested opening [."""
 
         pattern = bregex.compile_search(r'test [[] \R', regex.UNICODE)
-        self.assertEqual(pattern.pattern, r'test [[] (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)', regex.UNICODE)
+        self.assertEqual(pattern.pattern, r'test [[] (?>\r\n|[\n\v\f\r\x85\u2028\u2029])', regex.UNICODE)
 
     def test_posix_parse(self):
         """Test posix in a group."""
 
         pattern = bregex.compile_search(r'test [[:graph:]] \R', regex.V0)
-        self.assertEqual(pattern.pattern, r'test [[:graph:]] (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)')
+        self.assertEqual(pattern.pattern, r'test [[:graph:]] (?>\r\n|[\n\v\f\r\x85\u2028\u2029])')
 
         pattern = bregex.compile_search(r'test [[:graph:]] \R', regex.V1)
-        self.assertEqual(pattern.pattern, r'test [[:graph:]] (?>\r\n|\n|\x0b|\f|\r|\x85|\u2028|\u2029)')
+        self.assertEqual(pattern.pattern, r'test [[:graph:]] (?>\r\n|[\n\v\f\r\x85\u2028\u2029])')
 
     def test_inline_comments(self):
         """Test that we properly find inline comments and avoid them."""
@@ -577,6 +594,12 @@ class TestSearchTemplate(unittest.TestCase):
         m = pattern2.match('test')
         self.assertTrue(m is not None)
 
+    @unittest.skipUnless(PY39_PLUS, "Python 3.9 required")
+    def test_emoji_property(self):
+        """Test new emoji properties."""
+
+        self.assertEqual(len(bregex.findall(r'\p{emoji}', '\U0001F600 test \U0001F600')), 2)
+
 
 class TestReplaceTemplate(unittest.TestCase):
     """Test replace template."""
@@ -586,14 +609,17 @@ class TestReplaceTemplate(unittest.TestCase):
 
         p1 = bregex.compile('(test)')
         p2 = bregex.compile('(test)')
+        p3 = bregex.compile(b'(test)')
         r1 = p1.compile(r'\1')
         r2 = p1.compile(r'\1')
         r3 = p2.compile(r'\1')
         r4 = p2.compile(r'\1', bregex.FORMAT)
+        r5 = p3.compile(br'\1')
 
         self.assertTrue(r1 == r2)
         self.assertTrue(r2 == r3)
         self.assertTrue(r1 != r4)
+        self.assertTrue(r1 != r5)
 
         r5 = copy.copy(r1)
         self.assertTrue(r1 == r5)
@@ -602,10 +628,10 @@ class TestReplaceTemplate(unittest.TestCase):
     def test_format_failures(self):
         """Test format parsing failures."""
 
-        with pytest.raises(SyntaxError):
+        with pytest.raises(_regex_core.error):
             bregex.subf('test', r'{1.}', 'test', bregex.FORMAT)
 
-        with pytest.raises(SyntaxError):
+        with pytest.raises(IndexError):
             bregex.subf('test', r'{a.}', 'test', bregex.FORMAT)
 
         with pytest.raises(SyntaxError):
@@ -623,6 +649,57 @@ class TestReplaceTemplate(unittest.TestCase):
         with pytest.raises(SyntaxError):
             bregex.subf('test', r'test { test', 'test', bregex.FORMAT)
 
+        with pytest.raises(_regex_core.error):
+            bregex.subf(b'test', br'{1.}', b'test', bregex.FORMAT)
+
+        with pytest.raises(IndexError):
+            bregex.subf(b'test', br'{a.}', b'test', bregex.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(b'test', br'{1[}', b'test', bregex.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(b'test', br'{a[}', b'test', bregex.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(b'test', br'test } test', b'test', bregex.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(b'test', br'test {test', b'test', bregex.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(b'test', br'test { test', b'test', bregex.FORMAT)
+
+        with pytest.raises(TypeError):
+            bregex.subf(b'test', br'{[test]}', b'test', bregex.FORMAT)
+
+    def test_format_conversions(self):
+        """Test string format conversion paths."""
+
+        self.assertTrue(bregex.subf('test', r'{0.index}', 'test').startswith('<built-in method'))
+        self.assertEqual(bregex.subf('test', r'{0.__class__.__name__}', 'test'), 'str')
+        self.assertTrue(bregex.subf('test', r'{0.index!s}', 'test').startswith('<built-in method'))
+        self.assertEqual(bregex.subf('test', r'{0.__class__.__name__!s}', 'test'), 'str')
+        self.assertTrue(bregex.subf('test', r'{0.index!a}', 'test').startswith('<built-in method'))
+
+        self.assertTrue(bregex.subf(b'test', br'{0.index}', b'test').startswith(b'<built-in method'))
+        self.assertEqual(bregex.subf(b'test', br'{0.__class__.__name__}', b'test'), b'bytes')
+        self.assertTrue(bregex.subf(b'test', br'{0.index!s}', b'test').startswith(b'<built-in method'))
+        self.assertEqual(bregex.subf(b'test', br'{0.__class__.__name__!s}', b'test'), b'bytes')
+        self.assertTrue(bregex.subf('test', r'{0.index!a}', 'test').startswith('<built-in method'))
+
+    def test_incompatible_strings(self):
+        """Test incompatible string types."""
+
+        with pytest.raises(TypeError):
+            bregex.compile('test').compile(b'test')
+
+        p1 = bregex.compile('test')
+        repl = bregex.compile(b'test').compile(b'other')
+        m = p1.match('test')
+        with pytest.raises(TypeError):
+            repl(m)
+
     def test_named_unicode_failures(self):
         """Test named Unicode failures."""
 
@@ -673,14 +750,10 @@ class TestReplaceTemplate(unittest.TestCase):
     def test_unicode_narrow_value(self):
         """Test Unicode narrow value."""
 
-        # Python 3 does not apply Unicode syntax in raw string,
-        # but Python 2 does.  So Python 2 will turn the following
-        # into a group and fail. There isn't really a workaround for this.
-        if PY3:
-            pattern = regex.compile(r"(some)(.+?)(pattern)(!)")
-            expand = bregex.compile_replace(pattern, r'\u005cg')
-            results = expand(pattern.match('some test pattern!'))
-            self.assertEqual('\g', results)
+        pattern = regex.compile(r"(some)(.+?)(pattern)(!)")
+        expand = bregex.compile_replace(pattern, r'\u005cg')
+        results = expand(pattern.match('some test pattern!'))
+        self.assertEqual(r'\g', results)
 
     def test_unexpected_end(self):
         """Test cases where there is an unexpected end to the replace string."""
@@ -703,8 +776,8 @@ class TestReplaceTemplate(unittest.TestCase):
             'line line line '
         )
 
-    def test_binary_line_break(self):
-        r"""Test binary line break \R."""
+    def test_bytes_line_break(self):
+        r"""Test bytes line break \R."""
 
         self.assertEqual(
             bregex.sub(br"\R", b' ', b'line\r\nline\nline\r'),
@@ -825,7 +898,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\c\l\2\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a test for Stacking!', results)
+        self.assertEqual('This is a test for stacking!', results)
 
     def test_span_stacked_case(self):
         """Test stacked casing of non-spans in and out of a span."""
@@ -835,7 +908,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\c\L\l\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a test for Stacking!', results)
+        self.assertEqual('This is a test for stacking!', results)
 
     def test_single_case_followed_by_bslash(self):
         """Test single backslash following a single case reference."""
@@ -865,7 +938,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'Test \l\Cstacked\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('Test sTACKED!', results)
+        self.assertEqual('Test STACKED!', results)
 
     def test_extraneous_end_char(self):
         """Test for extraneous end characters."""
@@ -955,7 +1028,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\C\C\2\E\3\E')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex UPPERCASE TEST!', results)
+        self.assertEqual('This is a complex UPPERCASE test!', results)
 
     def test_lower_after_lower(self):
         """Test lowercase after lowercase."""
@@ -975,7 +1048,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\L\L\2\E\3\E')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex lowercase test!', results)
+        self.assertEqual('This is a complex lowercase TEST!', results)
 
     def test_span_upper_after_lower(self):
         """Test lowercase followed by uppercase span."""
@@ -985,7 +1058,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\l\C\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex uPPERCASE test!', results)
+        self.assertEqual('This is a complex UPPERCASE test!', results)
 
     def test_span_lower_after_upper(self):
         """Test uppercase followed by lowercase span."""
@@ -995,7 +1068,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\1\c\L\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex Lowercase test!', results)
+        self.assertEqual('This is a complex lowercase test!', results)
 
     def test_span_upper_around_lower(self):
         """Test uppercase span around a lowercase."""
@@ -1099,10 +1172,10 @@ class TestReplaceTemplate(unittest.TestCase):
         results2 = pattern.sub(repl_pattern, text)
 
         self.assertEqual(results2, results)
-        self.assertEqual('\e \\e \\\e \\\\e \\\\\e', results)
+        self.assertEqual('\\e \\e \\\\e \\\\e \\\\\\e', results)
 
-    def test_binary_normal_escaping(self):
-        """Test binary normal escaped slash."""
+    def test_bytes_normal_escaping(self):
+        """Test bytes normal escaped slash."""
 
         text = b"This is a test of normal escaping!"
         pattern = regex.compile(br"(.+)")
@@ -1113,7 +1186,7 @@ class TestReplaceTemplate(unittest.TestCase):
         results2 = pattern.sub(repl_pattern, text)
 
         self.assertEqual(results2, results)
-        self.assertEqual(b'\e \\e \\\e \\\\e \\\\\e', results)
+        self.assertEqual(b'\\e \\e \\\\e \\\\e \\\\\\e', results)
 
     def test_escaped_slash_at_eol(self):
         """Test escaped slash at end of line."""
@@ -1159,7 +1232,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # Use uncompiled pattern when compiling replace.
         expand = bregex.compile_replace(pattern, r'\l\C\g<first>\l\g<second>\L\c\g<third>\E\g<fourth>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_mixed_groups2(self):
         """Test mix of upper and lower case with group indexes and a string replace pattern (2)."""
@@ -1171,7 +1244,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # This will pass because we do not need to resolve named groups.
         expand = bregex.compile_replace(pattern, r'\l\C\g<1>\l\g<2>\L\c\g<3>\E\g<4>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_mixed_groups3(self):
         """Test mix of upper and lower case with named groups and a compiled replace pattern (3)."""
@@ -1183,7 +1256,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # Now using compiled pattern, we can use named groups in replace template.
         expand = bregex.compile_replace(pattern, r'\l\C\g<first>\l\g<second>\L\c\g<third>\E\g<fourth>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_as_replace_function(self):
         """Test that replace can be used as a replace function."""
@@ -1197,27 +1270,16 @@ class TestReplaceTemplate(unittest.TestCase):
         self.assertEqual(result, "This is awesome!  Here we go!  This is awesome!  Here we go!")
         self.assertEqual(count, 2)
 
-    # def test_nested_group(self):
-    #     """Test that replace can be used as a replace function."""
+    def test_bytes_replace(self):
+        """Test that bytes regex result is a bytes string."""
 
-    #     text = "Here is a nested [] group search!"
-    #     text_pattern = r"(Here\ is\ a\ nested\ \[(?V1)\]\ group\ )([srch[aeiou]--[iou]]+)(?x)(!)"
-    #     pattern = bregex.compile_search(text_pattern)
-    #     expand = bregex.compile_replace(pattern, r'\1\2\3')
-    #     result = expand(pattern.match(text))
-
-    #     self.assertEqual(result, "Here is a nested [] group search!")
-
-    def test_binary_replace(self):
-        """Test that binary regex result is a binary string."""
-
-        text = b"This is some binary text!"
-        pattern = bregex.compile_search(br"This is (some binary text)!")
+        text = b"This is some bytes text!"
+        pattern = bregex.compile_search(br"This is (some bytes text)!")
         expand = bregex.compile_replace(pattern, br'\C\1\E')
         m = pattern.match(text)
         result = expand(m)
-        self.assertEqual(result, b"SOME BINARY TEXT")
-        self.assertTrue(isinstance(result, binary_type))
+        self.assertEqual(result, b"SOME BYTES TEXT")
+        self.assertTrue(isinstance(result, bytes))
 
     def test_template_replace(self):
         """Test replace by passing in replace template."""
@@ -1242,7 +1304,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(pattern, r'\l\C\g<0001>\l\g<02>\L\c\g<03>\E\g<004>\E\5\n\C\g<000>\E')
         results = expand(pattern.match(text))
         self.assertEqual(
-            'tHIS iS A TEST FOR Numeric capture GROUPS!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
             results
         )
 
@@ -1253,10 +1315,21 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(numeric capture )(groups)(!)"
         pattern = regex.compile(text_pattern)
 
-        expand = bregex.compile_replace(pattern, r'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bregex.FORMAT)
-        results = expand(pattern.match(text))
+        expandf = bregex.compile_replace(pattern, r'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bregex.FORMAT)
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            'tHIS iS A TEST FOR Numeric capture GROUPS!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            results
+        )
+
+        text = b"this is a test for numeric capture groups!"
+        text_pattern = br"(this )(.+?)(numeric capture )(groups)(!)"
+        pattern = regex.compile(text_pattern)
+
+        expandf = bregex.compile_replace(pattern, br'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bregex.FORMAT)
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
             results
         )
 
@@ -1267,12 +1340,25 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(format capture )(groups)(!)"
         pattern = regex.compile(text_pattern)
 
-        expand = bregex.compile_replace(
+        expandf = bregex.compile_replace(
             pattern, r'\l\C{{0001}}\l{{{02}}}\L\c{03}\E{004}\E{5}\n\C{000}\E', bregex.FORMAT
         )
-        results = expand(pattern.match(text))
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            '{0001}{IS A TEST FOR }Format capture GROUPS!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
+            '{0001}{IS A TEST FOR }Format capture groups!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
+            results
+        )
+
+        text = b"this is a test for format capture groups!"
+        text_pattern = br"(this )(.+?)(format capture )(groups)(!)"
+        pattern = regex.compile(text_pattern)
+
+        expandf = bregex.compile_replace(
+            pattern, br'\l\C{{0001}}\l{{{02}}}\L\c{03}\E{004}\E{5}\n\C{000}\E', bregex.FORMAT
+        )
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'{0001}{IS A TEST FOR }Format capture groups!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
             results
         )
 
@@ -1283,12 +1369,25 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(format capture )(groups)(!)"
         pattern = regex.compile(text_pattern)
 
-        expand = bregex.compile_replace(
+        expandf = bregex.compile_replace(
             pattern, r'\C{}\E\n\l\C{}\l{}\L\c{}\E{}\E{}{{}}', bregex.FORMAT
         )
-        results = expand(pattern.match(text))
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\ntHIS iS A TEST FOR Format capture GROUPS!{}',
+            'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\nTHIS iS A TEST FOR Format capture groups!{}',
+            results
+        )
+
+        text = b"this is a test for format capture groups!"
+        text_pattern = br"(this )(.+?)(format capture )(groups)(!)"
+        pattern = regex.compile(text_pattern)
+
+        expandf = bregex.compile_replace(
+            pattern, br'\C{}\E\n\l\C{}\l{}\L\c{}\E{}\E{}{{}}', bregex.FORMAT
+        )
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\nTHIS iS A TEST FOR Format capture groups!{}',
             results
         )
 
@@ -1340,7 +1439,7 @@ class TestReplaceTemplate(unittest.TestCase):
             results
         )
 
-    def test_binary_format_capture_bases(self):
+    def test_bytes_format_capture_bases(self):
         """Test capture bases."""
 
         text = b"abababab"
@@ -1350,6 +1449,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bregex.compile_replace(
             pattern, br'{1[-0x1]}{1[0o3]}{1[0b101]}', bregex.FORMAT
         )
+
         results = expand(pattern.match(text))
         self.assertEqual(
             b'bbb',
@@ -1395,6 +1495,99 @@ class TestReplaceTemplate(unittest.TestCase):
         result = pattern.sub('\\C\\U00000070\\U0001F360\\E', 'Test')
         self.assertEqual(result, 'P\U0001F360')
 
+    def test_format_inter_escape(self):
+        """Test escaped characters inside format group."""
+
+        self.assertEqual(
+            bregex.subf(
+                r'(Te)(st)(ing)(!)',
+                r'\x7b1\x7d-\u007b2\u007d-\U0000007b3\U0000007d-\N{LEFT CURLY BRACKET}4\N{RIGHT CURLY BRACKET}',
+                'Testing!'
+            ),
+            "Te-st-ing-!"
+        )
+        self.assertEqual(
+            bregex.subf(
+                r'(Te)(st)(ing)(!)',
+                r'\1731\175-{2:\\^6}',
+                'Testing!'
+            ),
+            "Te-\\\\st\\\\"
+        )
+
+        with pytest.raises(SyntaxError):
+            bregex.subf(r'(test)', r'{\g}', 'test')
+
+        with pytest.raises(ValueError):
+            bregex.subf(br'(test)', br'{\777}', b'test')
+
+    def test_format_features(self):
+        """Test format features."""
+
+        pattern = bregex.compile(r'(Te)(st)(?P<group>ing)')
+        self.assertEqual(pattern.subf(r'{.__class__.__name__}', 'Testing'), 'str')
+        self.assertEqual(pattern.subf(r'{1:<30}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:30}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:>30}', 'Testing'), '                            Te')
+        self.assertEqual(pattern.subf(r'{1:^30}', 'Testing'), '              Te              ')
+        self.assertEqual(pattern.subf(r'{1:*^30}', 'Testing'), '**************Te**************')
+        self.assertEqual(pattern.subf(r'{1:^030}', 'Testing'), '00000000000000Te00000000000000')
+        self.assertEqual(pattern.subf(r'{1:^^30}', 'Testing'), '^^^^^^^^^^^^^^Te^^^^^^^^^^^^^^')
+        self.assertEqual(pattern.subf(r'{1:1^30}', 'Testing'), '11111111111111Te11111111111111')
+        self.assertEqual(pattern.subf(r'{1:<30s}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:s}', 'Testing'), 'Te')
+        self.assertEqual(pattern.subf(r'{2!r}', 'Testing'), "'st'")
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2!x}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{a$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{:ss}', 'Testing')
+
+        with pytest.raises(ValueError):
+            pattern.subf(r'{:030}', 'Testing')
+
+        pattern = bregex.compile(br'(Te)(st)(?P<group>ing)')
+        self.assertEqual(pattern.subf(br'{.__class__.__name__}', b'Testing'), b'bytes')
+        self.assertEqual(pattern.subf(br'{1:<30}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:30}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:>30}', b'Testing'), b'                            Te')
+        self.assertEqual(pattern.subf(br'{1:^30}', b'Testing'), b'              Te              ')
+        self.assertEqual(pattern.subf(br'{1:*^30}', b'Testing'), b'**************Te**************')
+        self.assertEqual(pattern.subf(br'{1:^030}', b'Testing'), b'00000000000000Te00000000000000')
+        self.assertEqual(pattern.subf(br'{1:^^30}', b'Testing'), b'^^^^^^^^^^^^^^Te^^^^^^^^^^^^^^')
+        self.assertEqual(pattern.subf(br'{1:1^30}', b'Testing'), b'11111111111111Te11111111111111')
+        self.assertEqual(pattern.subf(br'{1:<30s}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:s}', b'Testing'), b'Te')
+        self.assertEqual(pattern.subf(br'{2!r}', b'Testing'), b"b'st'")
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2!x}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{a$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{:ss}', b'Testing')
+
+        with pytest.raises(ValueError):
+            pattern.subf(br'{:030}', b'Testing')
+
     def test_dont_case_special_refs(self):
         """Test that we don't case Unicode and bytes tokens, but case the character."""
 
@@ -1418,7 +1611,7 @@ class TestReplaceTemplate(unittest.TestCase):
         results = expandf(pattern.match('Test'))
         self.assertEqual('\u0108\nWw\u0108', results)
 
-        # Binary doesn't care about Unicode, but should evaluate bytes
+        # Bytes doesn't care about Unicode, but should evaluate bytes
         pattern = regex.compile(b'Test')
         expand = bregex.compile_replace(pattern, br'\C\u0109\n\x77\E\l\x57\c\u0109')
         results = expand(pattern.match(b'Test'))
@@ -1455,8 +1648,8 @@ class TestReplaceTemplate(unittest.TestCase):
         self.assertEqual('W\u01b6W\u01b5\nw\u01b5', results)
 
         pattern = regex.compile(b'Test')
-        expandf = bregex.compile_replace(pattern, br'\127\C\167\n\E\l\127\c')
-        results = expandf(pattern.match(b'Test'))
+        expand = bregex.compile_replace(pattern, br'\127\C\167\n\E\l\127\c')
+        results = expand(pattern.match(b'Test'))
         self.assertEqual(b'WW\nw', results)
 
         # Null should pass through
@@ -1471,12 +1664,12 @@ class TestReplaceTemplate(unittest.TestCase):
         self.assertEqual(b'\x00\x00\x00', results)
 
         pattern = regex.compile('Test')
-        expand = bregex.compile_replace(pattern, r'\0\00\000', bregex.FORMAT)
-        results = expand(pattern.match('Test'))
+        expandf = bregex.compile_replace(pattern, r'\0\00\000', bregex.FORMAT)
+        results = expandf(pattern.match('Test'))
         self.assertEqual('\x00\x00\x00', results)
 
         pattern = regex.compile(b'Test')
-        expand = bregex.compile_replace(pattern, br'\0\00\000', bregex.FORMAT)
+        expandf = bregex.compile_replace(pattern, br'\0\00\000', bregex.FORMAT)
         results = expand(pattern.match(b'Test'))
         self.assertEqual(b'\x00\x00\x00', results)
 
@@ -1484,22 +1677,6 @@ class TestReplaceTemplate(unittest.TestCase):
 class TestExceptions(unittest.TestCase):
     """Test Exceptions."""
 
-    # def test_incomplete_replace_narrow_unicode(self):
-    #     """Test incomplete replace of narrow Unicode."""
-
-    #     p = bregex.compile_search(r'test')
-    #     with self.assertRaises(SyntaxError) as e:
-    #         bregex.compile_replace(p, r'Replace \u fail!')
-    #     self.assertTrue(str(e), 'Format for Unicode is \\uXXXX!')
-
-    # def test_incomplete_replace_wide_unicode(self):
-    #     """Test incomplete replace wide Unicode."""
-
-    #     p = bregex.compile_search(r'test')
-    #     with self.assertRaises(SyntaxError) as e:
-    #         bregex.compile_replace(p, r'Replace \U fail!')
-    #     self.assertTrue(str(e), 'Format for wide Unicode is \\UXXXXXXXX!')
-
     def test_immutable(self):
         """Test immutable object."""
 
@@ -1547,13 +1724,11 @@ class TestExceptions(unittest.TestCase):
         text_pattern = r"(\w)+"
         pattern = regex.compile(text_pattern)
 
-        with pytest.raises(ValueError) as excinfo:
-            bregex.compile_replace(
-                pattern, r'{1[0o3f]}', bregex.FORMAT
+        with pytest.raises(TypeError):
+            bregex.subf(
+                pattern, r'{1[0o3f]}', 'test'
             )
 
-        assert "Capture index must be an integer!" in str(excinfo.value)
-
     def test_format_bad_capture_range(self):
         """Test a bad capture."""
 
@@ -1563,11 +1738,9 @@ class TestExceptions(unittest.TestCase):
             pattern, r'{1[37]}', bregex.FORMAT
         )
 
-        with pytest.raises(IndexError) as excinfo:
+        with pytest.raises(IndexError):
             expand(pattern.match('text'))
 
-        assert "is out of range!" in str(excinfo.value)
-
     def test_require_compiled_pattern(self):
         """Test a bad capture."""
 
@@ -1751,6 +1924,29 @@ class TestExceptions(unittest.TestCase):
 
         assert "octal escape value outside of range 0-0o377!" in str(excinfo.value)
 
+    @unittest.skipUnless(PY36_PLUS, "Ignore in PY 3.5")
+    def test_timeout(self):
+        """Test timeout."""
+
+        def fast_replace(m):
+            """Fast replace."""
+            return 'X'
+
+        def slow_replace(m):
+            start = time.time()
+            while True:
+                elapsed = time.time() - start
+                if elapsed >= 0.5:
+                    break
+            return 'X'
+
+        self.assertEqual(regex.sub(r'[a-z]', fast_replace, 'abcdef', timeout=2), 'XXXXXX')
+
+        with pytest.raises(TimeoutError) as excinfo:
+            regex.sub(r'[a-z]', slow_replace, 'abcdef', timeout=2)
+
+        assert "regex timed out" in str(excinfo.value)
+
 
 class TestConvenienceFunctions(unittest.TestCase):
     """Test convenience functions."""
